/*  FILENAME: boxobj.c

    AUTHOR:  Kelvin Thompson, 10/87

    DESCRIPTION:  Routines for handling the object(s) displayed in 
      the viewbox.
	You can modify this file to have 'boxview' display objects
      of your own creation.  Just write a new version of 'init_obj'
      (which 'boxview' calls shortly after startup) and 'draw_obj'
      (which 'boxview' calls every time it wants to draw the object).
      See the function code itself for more information.
*/

#include <stdio.h>
#include <math.h>
#include <gl.h>
#include <device.h>
#include "ktypes.h"
#include "idraw.h"

/* programs in this file */
extern void init_obj();
extern void draw_obj();
extern void gen_sphere();
extern int gen_geomobj();

/* globals */
static Object theobj;




/*  ************************************
    *                                  *
    *       init_obj                   *
    *                                  *
    ************************************

    DESCRIPTION:  Initialize the object to be shown.
	This routine is called once, shortly after 'boxview'
      starts up.  Graphics operations have been initialized,
      but the user has not yet seen anything on the screen.
	This routine should do its initializations based only
      on input provided by the entry parameters 'argc' and 'argv'.
      This routine should not ask the user for information.
	Programmers should use calls of the form 

             cm_mapcolor( seg1, s1_OBJSTART+idx, R, G, B );

      to define color map entries.  'seg1' and 's1_OBJSTART' are
      globals;  idx varies for different index values;  R,G,B are
      the color entries in the table.  Corresponding references to
      the color map should take the form

             color( s1_OBJSTART+idx );

    ENTRY:
      argc,argv -- Standard Unix input.
*/

void init_obj ( argc, argv )
int argc;
char **argv;
{
FILE *geomfile;
int got_obj;

/* init */
theobj = genobj();
got_obj = FALSE;

makeobj( theobj );

  while ( --argc > 0 )
    {
    argv++;
    got_obj |= gen_geomobj( *argv );
    }
  
  if ( ! got_obj )
    gen_sphere();

closeobj();
}




/*  ************************************
    *                                  *
    *       draw_obj                   *
    *                                  *
    ************************************

    DESCRIPTION:  Draw the object.
	This routine is called every time 'boxview' wants to redraw
      the object it is showing.  It is called every refresh time
      (up to 60 times a second) even if the user performs no action,
      so moving objects can be displayed in this routine.
	The drawn object should fit inside the cube from (-1,-1,-1) to
      (1,1,1).  Only drawing and model-transforming calls should be made
      inside this routine (i.e. no viewing or windowing calls).
*/

void draw_obj ( )
{
pushmatrix();
  callobj( theobj );
popmatrix();
}




/*  ************************************
    *                                  *
    *       gen_sphere                 *
    *                                  *
    ************************************

    DESCRIPTION:  Generate a colorful sphere-like object.  This is
      intended to be a cheap default object, in case the real object
      can't be obtained.
*/

void gen_sphere ( )
{
/* three colors in sphere */
cm_mapcolor( seg1, s1_OBJSTART+0, 255, 100, 100 );
cm_mapcolor( seg1, s1_OBJSTART+1, 100, 255, 100 );
cm_mapcolor( seg1, s1_OBJSTART+2, 100, 100, 255 );

color( s1_OBJSTART+0 );
circ( 0.0, 0.0, 0.5 );
rotate( 90*10, 'x' );
color( s1_OBJSTART+1 );
circ( 0.0, 0.0, 0.5 );
rotate( 90*10, 'y' );
color( s1_OBJSTART+2 );
circ( 0.0, 0.0, 0.5 );
}




/*  ************************************
    *                                  *
    *       gen_geomobj                *
    *                                  *
    ************************************

    DESCRIPTION:  Load an object consisting of polygons from a file
      and store it in an Iris display list.  The object will render
      as a wireframe.

    ENTRY:  Filename of a geometry file.

    EXIT:  Returns FALSE if something went wrong, TRUE otherwise.
*/

#define MAX_POLY_EDGES 10

int gen_geomobj ( filename )
char *filename;
{
register i,j,temp;
int vcount,pcount,ecount,vindex;
register VEC3 *vertex,*vptr;
register float thisdist;
VEC3 thispoly[MAX_POLY_EDGES];
FILE *gstream;
static float maxdist=0.0;
static int firstone=TRUE;

/* open the file */
gstream = fopen( filename, "r" );
if ( gstream == 0 )
  return FALSE;

/* read top line */
temp = fscanf( gstream, "%d %d %d", &vcount, &pcount, &ecount );
if ( temp != 3 )
  { fclose(gstream); return FALSE; }

/* allocate for vertex array */
vertex = (VEC3 *) malloc( vcount * sizeof(VEC3) );
if ( vertex == 0 )
  { fclose(gstream); return FALSE; }

/* read in vertex array */
for ( 
    i = 0,        vptr = vertex; 
    i < vcount; 
    i++,          vptr++
    )
  {
  /* read in the point */
  temp = fscanf( gstream, "%f %f %f", &vptr->x, &vptr->y, &vptr->z );
  if ( temp != 3 )
    { free(vertex);  fclose(gstream);  return FALSE; }

  /* update maximum distance */
  if ( firstone )
    {
    thisdist = vptr->x * vptr->x + vptr->y * vptr->y + vptr->z * vptr->z;
    if ( thisdist > maxdist )
      maxdist = thisdist;
    }
  }

/* scale vertex array */
if ( firstone )
  maxdist = 1.0 / sqrt(maxdist);
for ( 
    i = 0,        vptr = vertex; 
    i < vcount; 
    i++,          vptr++
    )
  {
  vptr->x *= maxdist;
  vptr->y *= maxdist;
  vptr->z *= maxdist;
  }

/* define and set object color */
cm_mapcolor( seg1, s1_OBJSTART, 255, 100, 100 );
color( s1_OBJSTART );

/* read in polygons */
for ( i=0; i<pcount; i++ )
  {
  /* get edges in polygon */
  temp = fscanf( gstream, "%d", &ecount );
  if ( temp != 1 )
    { free(vertex);  fclose(gstream);  return FALSE; }

  /* loop through edges for this polygon */
  for ( j=0; j<ecount; j++ )
    {
    temp = fscanf( gstream, "%d", &vindex );
    if ( temp != 1 )
      { free(vertex);  fclose(gstream);  return FALSE; }
    vindex--;
    if ( j < MAX_POLY_EDGES )
      thispoly[j] = vertex[vindex];
    }

  /* send out polygon */
  if ( ecount <= MAX_POLY_EDGES )
    poly( ecount, thispoly );
  }

/* close and finish */
free( vertex );
fclose( gstream );
firstone = FALSE;
return TRUE;
}
