/*  FILENAME:  catgeom.c

    DESCRIPTION:  Concatenate several geom files into one.

    AUTHOR:  Kelvin Thompson, 4/25/89
*/

#include <stdio.h>
#include <math.h>

/* programs in this file */
extern main();
extern struct SHELL *read_gfile();
extern void write_catgeom();

#ifndef NULL
#  define NULL 0
#endif

#ifndef TRUE
#  define TRUE 1
#  define FALSE 0
#endif

typedef double FLOAT;
typedef int INT;
typedef int BOOL;

typedef struct VEC3 { FLOAT x,y,z; } VEC3;

typedef struct POLYGON
  {
  int npoints;
  VEC3 **point;
  } POLYGON;

typedef struct SHELL  /* not a Unix-shell; a 3D object-shell */
  {
  struct SHELL *next;	/* link to next in list */
  int npoints;
  int npolys;
  int nedges;
  VEC3 *point;		/* array of 'npoints' VEC3 structs */
  POLYGON *poly;	/* array of 'npolys' POLYGON structs */
  } SHELL;




/*  ************************************
    *                                  *
    *       main                       *
    *                                  *
    ************************************

    DESCRIPTION:  Entry point for catgeom utility.

    ENTRY:  Standard Unix.  Expects a list of .geom files to be
      (effectively) concatenated together.

    EXIT:  Standard Unix.  Returns zero on normal completion, 1
      on error.  Most errors involve problems with opening files
      or inability to allocate memory.
*/

main ( argc, argv )
int argc;
char **argv;
{
int fileidx;
char *filename;
FILE *fptr;
SHELL *gobj,*prevgobj;
int gobj_count;
SHELL **garray;
BOOL negpoly;

/* init */
gobj = NULL;
gobj_count = 0;

for ( fileidx=1; fileidx<argc; ++fileidx )
  {
  /* get filename */
  filename = argv[fileidx];

  /* check for minus-flag and stdin */
  if ( filename[0] != '-' )
    { fptr = NULL;  negpoly = FALSE; }
  else
    {
    if ( filename[1] == '\0' )
      { negpoly = FALSE;  fptr = stdin; }
    else if ( filename[1] == '-'  &&  filename[2] == '\0' )
      { negpoly = TRUE;  fptr = stdin; }
    else
      { negpoly = TRUE;  fptr = NULL;  filename++; }
    }

  /* open file */
  if ( fptr == NULL )
    {
    fptr = fopen( filename, "r" );
    if ( fptr == NULL )
      {
      fprintf( stderr, "Couldn't open file '%s'.\n", filename );
      perror( "catgeom" );
      exit( 1 );
      }
    }

  /* read in object and link up */
  prevgobj = gobj;
  gobj = read_gfile( fptr, negpoly );
  gobj->next = prevgobj;
  gobj_count++;

  /* finish up with this file */
  if ( fptr != stdin )
    fclose( fptr );
  }

/* get space for array of pointers */
garray = (SHELL **) malloc( gobj_count * sizeof(SHELL *) );
if ( garray == NULL )
  {
  fprintf( stderr, "Can't alloc enough memory.\n" );
  exit( 1 );
  }

/* fill array of pointers */
for (
			fileidx = gobj_count - 1;
    gobj != NULL;
    gobj = gobj->next,  --fileidx
    )
  garray[fileidx] = gobj;

write_catgeom( stdout, gobj_count, garray );
exit( 0 );
}




/*  *******************************
    *                             *
    *      read_gfile             *
    *                             *
    *******************************

    DESCRIPTION:  Read a .geom file and fill in a SHELL struct
      accordingly.

    ENTRY:
      fptr -- pointer to already-opened readable stream
      reverse -- boolean set to TRUE if handedness of polygons
	should be reversed; FALSE otherwise
*/

SHELL *read_gfile ( fptr, reverse )
FILE *fptr;
BOOL reverse;
{
register i,j;
register SHELL *obj;
VEC3 *pt;
POLYGON *poly;
int pointcount,polycount,vcount;
FLOAT ftemp,maxr;
int itemp;

/* get point and polygon counts (ignore edge count) */
fscanf( fptr, "%d %d %d", &pointcount, &polycount, &itemp );

/* allocate space */
obj = (SHELL *) malloc( sizeof(SHELL)
         + pointcount * sizeof(VEC3)
         + polycount * sizeof(POLYGON) );
if ( obj == NULL )
  {
  fprintf( stderr, "xgeom: Can't allocate space.\n" );
  exit( 1 );
  }

/* init object struct */
obj->npoints = pointcount;
obj->npolys = polycount;
obj->nedges = itemp;
obj->point = (VEC3 *) (& obj[1]);
obj->poly = (POLYGON *) (& obj->point[pointcount]);

/* read in points */
maxr = 0.0;
for (
    i = pointcount,   pt = obj->point;
    i;
    --i,              ++pt
    )
  {
  fscanf( fptr, "%lf %lf %lf", &pt->x, &pt->y, &pt->z );

  ftemp = fabs(pt->x);  if ( maxr < ftemp )  maxr = ftemp;
  ftemp = fabs(pt->y);  if ( maxr < ftemp )  maxr = ftemp;
  ftemp = fabs(pt->z);  if ( maxr < ftemp )  maxr = ftemp;
  }

/* read in polygons */
for (
    i = polycount,   poly = obj->poly;
    i;
    --i,              ++poly
    )
  {
  fscanf( fptr, "%d", &vcount );
  poly->npoints = vcount;

  poly->point = (VEC3 **) malloc( vcount * sizeof(VEC3 *) );
  if ( poly->point == NULL )
    {
    fprintf( stderr, "xgeom: Can't alloc tiny space.\n" );
    exit( 1 );
    }

  if ( reverse )
    {
    for ( j=vcount-1;  j>=0;  --j )
      {
      fscanf( fptr, "%d", &itemp );
      poly->point[j] = & obj->point[itemp-1];
      }
    }
  else
    {
    for ( j=0;  j<vcount;  ++j )
      {
      fscanf( fptr, "%d", &itemp );
      poly->point[j] = & obj->point[itemp-1];
      }
    }
  }

return obj;
}




/*  ************************************
    *                                  *
    *         write_catgeom            *
    *                                  *
    ************************************

    DESCRIPTION:

    ENTRY:
*/

void write_catgeom ( fptr, gcount, gobj )
FILE *fptr;
int gcount;
SHELL **gobj;
{
register VEC3 *pt;
register POLYGON *poly;
int gidx,ptidx,pyidx;
SHELL *thisg;
int sumpoints,sumpolys,sumedges;
int *ptoff,totloff;

/* alloc array of offsets */
ptoff = (int *) malloc( gcount * sizeof(int *) );
if ( ptoff == NULL )
  {
  printf( stderr, "Can't malloc space.\n" );
  exit( 1 );
  }

/* get total number of points, polys, edges */
sumpoints = 0;
sumpolys = 0;
sumedges = 0;
for ( gidx=0; gidx<gcount; gidx++ )
  {
  ptoff[gidx] = sumpoints;

  thisg = gobj[gidx];
  sumpoints += thisg->npoints;
  sumpolys += thisg->npolys;
  sumedges += thisg->nedges;
  }
fprintf( fptr, "%d %d %d\n", sumpoints, sumpolys, sumedges );

/* write out points */
for ( gidx=0; gidx<gcount; gidx++ )
  {
  thisg = gobj[gidx];
  for (
      ptidx = thisg->npoints,   pt = thisg->point;
      ptidx;
      --ptidx,                  ++pt
      )
    {
    fprintf( fptr, "%22lg %22lg %22lg\n", pt->x, pt->y, pt->z );
    }
  }

/* write out polys */
for ( gidx=0; gidx<gcount; gidx++ )
  {
  thisg = gobj[gidx];
  totloff = ptoff[gidx] + 1;

  /* loop through polys */
  for (
      pyidx = thisg->npolys,   poly = thisg->poly;
      pyidx;
      --pyidx,                 ++poly
      )
    {
    fprintf( fptr, "%d", poly->npoints );

    /* loop through this poly's points */
    for ( ptidx = 0;  ptidx < poly->npoints;  ++ptidx )
      {
      fprintf( fptr, " %d", (poly->point[ptidx] - thisg->point) + totloff );
      }
    putc( '\n', fptr );
    }
  }
}
