/*  FILENAME:  cm.c

    DESCRIPTION:  Programs for allocating different sets of bitplanes
      and assigning color.  Assumes double buffering.
*/

#include <stdio.h>
#include <gl.h>
#include <device.h>

/* programs in this file */
extern cm_reset();
extern cm_avail();
extern unsigned cm_alloc();
extern cm_mapcolor();
extern cm_color();
extern Colorindex cm_truecolor();

typedef struct SEG
  {
  unsigned char  start;    /* first bitplane in segment */
  unsigned char  depth;    /* number of bitplanes in segment */
  unsigned short  baseidx; /* base global index */
  unsigned short  maxidx;  /* maximum segment color index */
  unsigned short  mask;    /* mask in global index space */
  } SEG;

static SEG cmseg[64]={ {0,0} };  /* array of segment descriptions */
static unsigned segcount=0;      /* number of segments allocated */



/*  ************************************
    *                                  *
    *           cm_reset               *
    *                                  *
    ************************************

    DESCRIPTION:  Resets the color map and associated allocations.

    EXIT:  Returns number of available bitplanes.
*/

cm_reset ( )
{
/* reset the color map */
mapcolor( 0, 0x0000, 0x0000, 0x0000 );
mapcolor( 1, 0xffff, 0x0000, 0x0000 );
mapcolor( 2, 0x0000, 0xffff, 0x0000 );
mapcolor( 3, 0xffff, 0xffff, 0x0000 );
mapcolor( 4, 0x0000, 0x0000, 0xffff );
mapcolor( 5, 0xffff, 0x0000, 0xffff );
mapcolor( 6, 0x0000, 0xffff, 0xffff );
mapcolor( 7, 0xffff, 0xffff, 0xffff );

/* reset allocations */
cmseg[0].depth = 0;
segcount = 0;
return getplanes();
}




/*  ************************************
    *                                  *
    *       cm_avail                   *
    *                                  *
    ************************************

    DESCRIPTION:  Gives number of unallocated bitplanes left.

    EXIT:  Number of unallocated bitplanes.
*/

cm_avail ( )
{
register unsigned total;
register SEG *sptr;

sptr = & cmseg[ segcount - 1 ];
total = sptr->start + sptr->depth;

return getplanes() - total;
}




/*  ************************************
    *                                  *
    *        cm_alloc                  *
    *                                  *
    ************************************

    DESCRIPTION:  Allocate a set of bitplanes.  Later
      segments occlude earlier segments.  Segments
      use bits starting in the low-order part of the word.

    ENTRY:
      bits -- Number of bits desired for segment.

    EXIT:  Returns two values stored in a single, 32-bit,
      unsigned word.  Upper 16 bits (masked with 0xffff0000)
      contain number of planes actually allocated.  Lower
      16 bits (masked with 0x0000ffff) contain segment id number
      (the index to the 'cmseg' array).  Returns 0 if no more
      bitplanes available.
*/

static unsigned short hibits[]=
  {
  0xffff, 0xfffe, 0xfffc, 0xfff8,
  0xfff0, 0xffe0, 0xffc0, 0xff80,
  0xff00, 0xfe00, 0xfc00, 0xf800,
  0xf000, 0xe000, 0xc000, 0x8000,
  0x0000
  };

unsigned cm_alloc ( bits )
register unsigned bits;
{
register unsigned total,avail;
register SEG *sptr;

/* see what's already been allocated */
if ( segcount != 0 )
  {
  sptr = & cmseg[ segcount - 1];
  total = sptr->start + sptr->depth;
  sptr++;
  }
else
  {
  sptr = cmseg;
  total = 0;
  }

/* quit if everything taken */
avail = getplanes() - total;
if ( avail == 0 )
  return 0;

/* fill the next SEG cell */
sptr->start = total;
sptr->depth = ( avail > bits ) ? bits : avail;
sptr->baseidx = (1 << sptr->start) - 1;
sptr->maxidx = (1 << sptr->depth) - 1;
sptr->mask = sptr->maxidx << sptr->start;
sptr->mask = hibits[ sptr->start ];

/* mark the last SEG cell */
sptr[1].depth = 0;

return (sptr->depth << 16) | segcount++;
}




/*  ************************************
    *                                  *
    *        cm_mapcolor               *
    *                                  *
    ************************************

    DESCRIPTION:  Allocate a color in a bitplane segment.

    ENTRY:
      segid -- ID of segment whose color is to be defined.
      coloridx -- Color index to be defined.
      red,grn,blu -- Color description.  Set red color
        to (int)(-1) or 0xffffffff for transparent color.

    EXIT:  Returns 0 if 'segid' or 'coloridx' was bogus; nonzero
      otherwise.
*/

cm_mapcolor ( segid, coloridx, red, grn, blu )
unsigned segid;
Colorindex color;
unsigned red,grn,blu;
{
register unsigned i,loopmax,newcolor;
register SEG *sptr;
short prevred,prevgrn,prevblu;

/* get segment cell */
if ( segid >= segcount )
  return 0;
sptr = & cmseg[segid];

/* check index */
if ( coloridx > sptr->maxidx )
  return 0;

/* define the color */
loopmax = sptr->baseidx;
newcolor = coloridx << sptr->start;
if ( red == 0xffffffff )
  {
  /* defining transparent color */
  for ( i=0; i<=loopmax; i++ )
    {
    getmcolor( i, &prevred, &prevgrn, &prevblu );
    mapcolor( newcolor, prevred, prevgrn, prevblu );
    newcolor++;
    }
  }
else
  {
  /* defining an occluding color */
  for ( i=0; i<=loopmax; i++ )
    {
    mapcolor( newcolor, red, grn, blu );
    newcolor++;
    }
  }

return 1;
}




/*  ************************************
    *                                  *
    *        cm_color                  *
    *                                  *
    ************************************

    DESCRIPTION:  Set the current color to a segment's color.
      Leaves other segments the same.

    ENTRY:
      segid -- ID of segment being referred to.
      coloridx -- Color index.

    EXIT:  Returns 0 if 'segid' is bogus; nonzero otherwise.
*/

cm_color ( segid, coloridx )
unsigned segid;
Colorindex coloridx;
{
register SEG *sptr;

/* fprintf( stderr, "Called 'cm_color' with %d, %d\n",
  segid, coloridx ); */

/* get segment cell */
if ( segid >= segcount )
  return 0;
sptr = & cmseg[segid];

/* mask planes and set color */
writemask( sptr->mask );
color( (coloridx & sptr->maxidx) << sptr->start );

/* fprintf( stderr, "Mask = 0x%x; color = 0x%x\n",
  sptr->mask, ((coloridx & sptr->maxidx) << sptr->start) ); */

return 1;
}



/*  ************************************
    *                                  *
    *        cm_truecolor              *
    *                                  *
    ************************************

    DESCRIPTION:  Return the true color associated with a
      segment color.  Also turns on all writemasks.

    ENTRY:
*/

Colorindex cm_truecolor ( segid, coloridx )
unsigned segid;
Colorindex coloridx;
{
register SEG *sptr;

/* get segment cell */
if ( segid >= segcount )
  return 0;
sptr = & cmseg[segid];

/* unmask planes and set color */
writemask( 0xffffffff );
return (coloridx & sptr->maxidx) << sptr->start;
}
