/*  FILENAME: manip.c

    DESCRIPTION:  Subroutines for user manipulations.
*/

#include <gl.h>
#include <device.h>
#include "idraw.h"

#if PIPE==P_GT  ||  PIPE==P_PERSONAL
#  include <gl/feed.h>
#endif

/* programs in this file */
extern void init_manip();
extern void make_boxes();
extern void draw_box();
extern check_pick();

/* local variables */
static Object box_obj;



/*  ************************************
    *                                  *
    *       init_manip                 *
    *                                  *
    ************************************

    DESCRIPTION:  Manipulation initializations.
*/

void init_manip ( )
{
make_boxes();
}



/*  ************************************
    *                                  *
    *        make_boxes                *
    *                                  *
    ************************************

    DESCRIPTION:  Make object(s) for the main manipulator box.
*/

void make_boxes ( )
{
/* get the object ID */
box_obj = genobj();

/* the drawn box with axes */
makeobj( box_obj );
  cm_color( seg0, s0_3WDW );      /* the extent box */
  move( -1.0, -1.0, -1.0 );
  draw( -1.0,  1.0, -1.0 );
  draw(  1.0,  1.0, -1.0 );
  draw(  1.0, -1.0, -1.0 );
  draw( -1.0, -1.0, -1.0 );
  move( -1.0, -1.0,  1.0 );
  draw( -1.0,  1.0,  1.0 );
  draw(  1.0,  1.0,  1.0 );
  draw(  1.0, -1.0,  1.0 );
  draw( -1.0, -1.0,  1.0 );
  move( -1.0, -1.0, -1.0 );
  draw( -1.0, -1.0,  1.0 );
  move( -1.0,  1.0, -1.0 );
  draw( -1.0,  1.0,  1.0 );
  move(  1.0, -1.0, -1.0 );
  draw(  1.0, -1.0,  1.0 );
  move(  1.0,  1.0, -1.0 );
  draw(  1.0,  1.0,  1.0 );
closeobj();
}




/*  ************************************
    *                                  *
    *        draw_box                  *
    *                                  *
    ************************************

    DESCRIPTION:  Draw the main manipulator box.
*/

void draw_box ( )
{
callobj( box_obj );
}




/*  ************************************
    *                                  *
    *        check_pick                *
    *                                  *
    ************************************

    DESCRIPTION:  See if we pick anything that is drawn.
      The basic procedure:

	[1]  Set the Iris' transform so that a small region around
	     the mouse cursor maps to the entire screen.  Hence, 
	     every graphics primitive that is not very near the 
	     cursor will be clipped by the geometry engine.

	[2]  Put the Iris in feedback mode.

	[3]  Generate primitives and passthrus for the geometry
	     pipeline.  Primitives will go to the feedback buffer
	     rather than the screen.

	[4]  Take the Iris out of feedback mode.

	[5]  Look through the feedback buffer for primitives.
	     Passthrus always go through the engine, but primitives
	     that are completely clipped away will not.

    ENTRY:
      wdw -- Pointer to WINDOW struct for phony-window we will be
	picking inside.
      xform -- Pointer to Matrix for drawing transform.
      generate -- Pointer to a function which will generate the
        objects to be matched.
      param -- Parameter passed to function 'generate'.

    EXIT: Returns last matched 'passthrough' code.

    PREREQUISITES:  Need to set viewport first.
*/

check_pick ( wdw, xform, generate, param )
WINDOW *wdw;
Matrix xform;
void (*generate)();
{
register long count;
register outcode;
register betwflag;
register FBTYPE *codeptr;
RECT mousebox;
static Matrix pickmatx=I_INIT;

/* get the box for the mouse */
if ( ! getmousebox( wdw, &mousebox ) )
  return 0;

/* figure the transform */
qmap_rects( pickmatx, &mousebox, &ndcrect );

/* generate stuff for the geometry engine */
pushmatrix();
  loadmatrix( pickmatx );
  multmatrix( xform );
  feedback( fb_buff, FB_MAX );
  (*generate)( param );
  count = endfeedback( fb_buff );
popmatrix();

/* Parse the feedback buffer:  Every time there is a non-passthru
** thing (probably a primitive) between two passthrus, record the
** passthru value of the *trailing* passthru in 'outcode'.  We
** ignore adjacent passthrus because associated primitives have been
** clipped.  At the end of the parse-loop, 'outcode' will contain
** the passthru value of the last matched (picked; unclipped) primitive.
** If no primitives are in the feedback buffer, then none were picked,
** and 'outcode' returns zero.
*/
outcode = 0;
betwflag = FALSE;

#if PIPE==P_CLASSIC
  for ( codeptr = fb_buff;  (outcode==0) && (count>0);  --count )
    switch ( *codeptr++ )
      {
      case 8:    /* passthrough */
        if ( betwflag )
          {
          outcode = *codeptr;
          betwflag = FALSE;
          }
        codeptr++;
        --count;
        break;
  
      case 16:  /* move command */
      case 17:  /* draw command */
      case 18:  /* pnt command */
      case 48:  /* pmv command */
      case 49:  /* pdr command */
      case 51:  /* pclos command */
        betwflag = TRUE;
        codeptr += 2;
        count -= 2;
        break;
  
      case 56:  /* xfpt command */
        betwflag = TRUE;
        codeptr += 4;
        count -= 4;
        break;
  
      default:
        betwflag = TRUE;
        break;
      }
#endif  /* end parse for classic pipe */

#if PIPE==P_GT
  for ( codeptr = fb_buff;  (outcode==0) && (count>0); ) 
    {
    switch ( codeptr[0] )
      {
      case FB_PASSTHROUGH:    /* passthrough */
        if ( betwflag )
          {
          outcode = codeptr[2];
          betwflag = FALSE;
          }
        break;
      case FB_CMOV:
      case FB_LINE:
      case FB_POINT:
      case FB_POLYGON:
        betwflag = TRUE;
        break;
      }
    count -= codeptr[1] + 2;
    codeptr += codeptr[1] + 2;
    }
#endif  /* end parse for GT pipe */

#if PIPE==P_PERSONAL
  for ( codeptr = fb_buff;  (outcode==0) && (count>0); ) 
    {
    if ( *codeptr == FB_PASSTHROUGH )
      {
      if ( betwflag )
        {
        outcode = codeptr[1];
        betwflag = FALSE;
        }
      count -= 2;
      codeptr += 2;
      }
    else if ( *codeptr == FB_POINT  ||  *codeptr == FB_MOVE
		||  *codeptr == FB_DRAW )
      {
      betwflag = TRUE;
      count -= 7;
      codeptr += 7;
      }
    else if ( *codeptr == FB_CMOV )
      {
      betwflag = TRUE;
      count -= 4;
      codeptr += 4;
      }
    else /* ZBUFFER, LINESTYLE, SETPATTERN, LINEWIDTH, LSREPEAT */
      {
      /* don't count these as primitives */
      count -= 2;
      codeptr += 2;
      }
    }
#endif  /* end parse for Personal Iris pipe */

return outcode;
}
