/*  FILENAME: rotate.c (version II)

    DESCRIPTION:  Various programs dealing with rotations.
*/

#include <math.h>
#include <gl.h>
#include <device.h>
#include "idraw.h"

#define DRAWSPLAT TRUE
#define DRAWCELL  FALSE

/* programs in this file */
extern void init_rot();
extern check_rot();
extern void pick_rot();
extern void st_axis();
extern void st_rotate();
extern void gen_splat();
extern void drawsplat();
extern void drawcell();
extern void draw_rot();
extern void make_wbox();
extern void gen_wbox();

/* programs in 'xform.c' */
extern void rot_x(),rot_y(),rot_z();

/* other programs */
extern void st_draw();

static Object wbox_obj;

typedef struct SPLATCELL
  {
  float  a,b;
  float  ang;
  } SPLATCELL;

/* tweakable parameters */
#define NUM_TESTS  45
#define SPLAT_ANG  (PI/5.3)
#define NEG_START  0.0
#define POS_START  0.0
#define RANGE      1.4

/* basic parameters */
#define NUM_AXES    3
#define NUM_CORNERS 4
#define NUM_DIRS    3

/* derived parameters */
#define NUM_SPLATS  (NUM_TESTS|1)  /* make it odd */
#define SPLAT_SEP   (SPLAT_ANG/(NUM_SPLATS-2.0))

/* corner constructs */
#define CORNER(a,b) (((a&1)?2:0) | ((b&1)?1:0))
static SPLATCELL cornsplat[NUM_CORNERS] =
  {
  { -1.0, -1.0, -0.75*PI },
  { -1.0,  1.0,  0.75*PI },
  {  1.0, -1.0, -0.25*PI },
  {  1.0,  1.0,  0.25*PI }
  };

/* direction symbols */
#define NO_DIR    0
#define POS_DIR   1
#define NEG_DIR   2

/* axis symbols */
#define X_AXIS    TOKCOD_X
#define Y_AXIS    TOKCOD_Y
#define Z_AXIS    TOKCOD_Z

/* splat arrays */
static SPLATCELL splat[NUM_CORNERS][NUM_DIRS][NUM_SPLATS];
static Object splatobj[NUM_AXES][NUM_CORNERS][NUM_DIRS];

/* state of rotate action */
typedef struct ROTSTATE
  {
  void    (*func)();
  float   ang;
  UNS8    axis;
  UNS8    corner;
  UNS8    dir;
  } ROTSTATE;
static ROTSTATE rstate;
static UNS16 trysplat=FALSE,trycell=FALSE;
static short lastpick;



/*  *********************
    *                   *
    *     init_rot      *
    *                   *
    *********************

    DESCRIPTION:  Initialize rotation stuff.
*/

void init_rot ( )
{
register corner,i,dir;
float relang,absang;
float aa,bb;

/* first splat for each direction and corner is zero offset angle */
for ( corner=0; corner<NUM_CORNERS; corner++ )
  for ( i=0; i<NUM_DIRS; i++ )
    {
    splat[corner][i][0] = cornsplat[corner];
    splat[corner][i][0].ang = 0.0;
    }

/* initialize rest of splat array */
for ( i=1; i<NUM_SPLATS; i++ )
  {
  /* neutral direction */
  relang = ((i&1)?-1:1) * ( (SPLAT_SEP*0.5) + ((i-1)>>1) * SPLAT_SEP );
  for ( corner=0; corner<NUM_CORNERS; corner++ )
    {
    absang = cornsplat[corner].ang + relang;
    splat[corner][NO_DIR][i].ang = relang;
    splat[corner][NO_DIR][i].a = SQRT_2 * cos(absang);
    splat[corner][NO_DIR][i].b = SQRT_2 * sin(absang);
    }

  /* negative direction */
  relang = NEG_START - (i-1) * SPLAT_SEP;
  for ( corner=0; corner<NUM_CORNERS; corner++ )
    {
    absang = cornsplat[corner].ang + relang;
    splat[corner][NEG_DIR][i].ang = relang;
    splat[corner][NEG_DIR][i].a = SQRT_2 * cos(absang);
    splat[corner][NEG_DIR][i].b = SQRT_2 * sin(absang);
    }

  /* positive direction */
  relang = POS_START + (i-1) * SPLAT_SEP;
  for ( corner=0; corner<NUM_CORNERS; corner++ )
    {
    absang = cornsplat[corner].ang + relang;
    splat[corner][POS_DIR][i].ang = relang;
    splat[corner][POS_DIR][i].a = SQRT_2 * cos(absang);
    splat[corner][POS_DIR][i].b = SQRT_2 * sin(absang);
    }
  }

/* initialize splat objects */
for ( corner=0; corner<NUM_CORNERS; corner++ )
  for ( dir=0; dir<NUM_DIRS; dir++ )
    {
    /* X axis */
    makeobj( splatobj[X_AXIS][corner][dir] = genobj() );
    for ( i=0; i<NUM_SPLATS; i++ )
      {
      aa = splat[corner][dir][i].a;
      bb = splat[corner][dir][i].b;
      move( -RANGE, aa, bb );
      draw(  RANGE, aa, bb );
      passthrough( i+1 );
      }
    closeobj();

    /* Y axis */
    makeobj( splatobj[Y_AXIS][corner][dir] = genobj() );
    for ( i=0; i<NUM_SPLATS; i++ )
      {
      aa = splat[corner][dir][i].a;
      bb = splat[corner][dir][i].b;
      move( bb, -RANGE, aa );
      draw( bb,  RANGE, aa );
      passthrough( i+1 );
      }
    closeobj();

    /* Z axis */
    makeobj( splatobj[Z_AXIS][corner][dir] = genobj() );
    for ( i=0; i<NUM_SPLATS; i++ )
      {
      aa = splat[corner][dir][i].a;
      bb = splat[corner][dir][i].b;
      move( aa, bb, -RANGE );
      draw( aa, bb,  RANGE );
      passthrough( i+1 );
      }
    closeobj();
    }

make_wbox();
}




/*  ************************************
    *                                  *
    *        check_rot                 *
    *                                  *
    ************************************

    DESCRIPTION:  Check rotation objects for a hit.

    EXIT:  Returns TRUE when we find something, FALSE otherwise.
*/

check_rot ( vptr )
register VIEW *vptr;
{
register UNS16 pickcode;
register WINDOW *wdw;
register out;

wdw = vptr->wdw;
viewport( wdw->edge.a.x, wdw->edge.b.x, wdw->edge.a.y, wdw->edge.b.y );
pickcode = check_pick( vptr->wdw, &vptr->total, gen_wbox );

switch ( pickcode & TOKTYP_MASK )
  {
  case TOKTYP_ROT:
    /* reactive rotation */
    pick_rot( pickcode & TOKCOD_MASK );
    draw_rot( vptr );
    out = TRUE;
    break;

  default:
    /* can't interpret -- punt */
    out = FALSE;
    break;
  }

return out;
}




/*  *********************
    *                   *
    *    pick_rot       *
    *                   *
    *********************

    DESCRIPTION:  Perform actions after picking a rotate bar.

    ENTRY:  Code for selected object.
*/

void pick_rot ( code )
UNS16 code;
{
static void (*rotfunc[NUM_AXES])() = { rot_x, rot_y, rot_z };

rstate.axis = code & TOKCOD_DIR;
rstate.dir = NO_DIR;
rstate.corner = (code & TOKPT(1,1)) >> 2;
rstate.func = rotfunc[rstate.axis];
rstate.ang = 0.0;

cursoff();
add_state( st_rotate, 0, 0 );
}



/*  ************************************
    *                                  *
    *        st_axis                   *
    *                                  *
    ************************************

    DESCRIPTION:  The axis rotation state.

    ENTRY:
      axis -- Axis of rotation.
      dum -- ignored dummy parameter
*/

void st_axis ( axis, dum )
UNS8 axis;
{
if ( !getbutton(LEFTMOUSE) )
  readystate();
else
  {
  switch ( axis )
    {
    case TOKCOD_X:
      rot_x( & v0ptr->box_to_rot, 2.0*PI/200.0 );
      break;

    case TOKCOD_Y:
      rot_y( & v0ptr->box_to_rot, 2.0*PI/200.0 );
      break;

    case TOKCOD_Z:
      rot_z( & v0ptr->box_to_rot, 2.0*PI/200.0 );
      break;
    }

  draw_rot( v0ptr );
  add_state( st_axis, axis, 0 );
  }
}




/*  *********************
    *                   *
    *    st_rotate      *
    *                   *
    *********************

    DESCRIPTION:  Reactive rotation

    ENTRY:
      dum1,dum2 -- ignored
*/

void st_rotate ( dum1, dum2 )
{
register i;
register SPLATCELL *splatptr;
register float ang;
UNS16 pickcode;

/* go back to waiting state if no mouse button */
if ( !getbutton(LEFTMOUSE) )
  readystate();

else
  {
  pickcode = check_pick( v0ptr->wdw, &v0ptr->total, gen_splat );

  if ( pickcode == 0 )
    {
    if ( rstate.dir != NO_DIR )
      (*rstate.func)( & v0ptr->box_to_rot, rstate.ang );
    rstate.dir = NO_DIR;
    }
  else
    {
#   if DRAWCELL
    trycell = TRUE;
    lastpick = pickcode;
#   endif

    rstate.ang = ang = splat[rstate.corner][rstate.dir][pickcode-1].ang;
    if ( ang == 0.0 )
      rstate.dir = NO_DIR;
    else
      {
      (*rstate.func)( & v0ptr->box_to_rot, ang );
      rstate.dir = (ang>0.0) ? POS_DIR : NEG_DIR;
      }
    }

  trysplat = TRUE;
  draw_rot( v0ptr );
  add_state( st_rotate, 0, 0 );
  trysplat = FALSE;
  trycell = FALSE;
  }
}




/*  ************************************
    *                                  *
    *          gen_splat               *
    *                                  *
    ************************************

    DESCRIPTION:

    ENTRY:
*/

void gen_splat ( )
{
callobj( splatobj[rstate.axis][rstate.corner][rstate.dir] );
}



#if DRAWSPLAT
/*  *********************
    *                   *
    *    drawsplat      *
    *                   *
    *********************

    DESCRIPTION:

    ENTRY:
*/

void drawsplat ( )
{
register i;
register SPLATCELL *splatptr;

cm_color( seg1, s1_SPLAT );
switch ( rstate.axis )
  {
  case X_AXIS:
    for (
      i = 0,      splatptr = splat[rstate.corner][rstate.dir];
      i < NUM_SPLATS;
      i++,      splatptr++
      )
      {
      move( -RANGE, splatptr->a, splatptr->b );
      draw(  RANGE, splatptr->a, splatptr->b );
      }
    break;

  case Y_AXIS:
    for (
      i = 0,      splatptr = splat[rstate.corner][rstate.dir];
      i < NUM_SPLATS;
      i++,      splatptr++
      )
      {
      move( splatptr->b, -RANGE, splatptr->a );
      draw( splatptr->b,  RANGE, splatptr->a );
      }
    break;

  case Z_AXIS:
    for (
      i = 0,      splatptr = splat[rstate.corner][rstate.dir];
      i < NUM_SPLATS;
      i++,      splatptr++
      )
      {
      move( splatptr->a, splatptr->b, -RANGE );
      draw( splatptr->a, splatptr->b,  RANGE );
      }
    break;
  }
}
#endif



#if DRAWCELL
/*  *********************
    *                   *
    *   drawcell        *
    *                   *
    *********************

    DESCRIPTION:

    ENTRY:
*/

void drawcell ( idx )
UNS16 idx;
{
register SPLATCELL *splatptr;

splatptr = & splat[rstate.corner][rstate.dir][idx];
finish();
cm_color( seg1, s1_SPLAT );
switch ( rstate.axis )
  {
  case X_AXIS:
    move( -RANGE, splatptr->a, splatptr->b );
    draw(  RANGE, splatptr->a, splatptr->b );
    break;

  case Y_AXIS:
    move( splatptr->b, -RANGE, splatptr->a );
    draw( splatptr->b,  RANGE, splatptr->a );
    break;

  case Z_AXIS:
    move( splatptr->a, splatptr->b, -RANGE );
    draw( splatptr->a, splatptr->b,  RANGE );
    break;
  }
}
#endif



/*  ************************************
    *                                  *
    *        make_wbox                 *
    *                                  *
    ************************************

    DESCRIPTION:

    ENTRY:
*/

void make_wbox ( )
{
wbox_obj = genobj();

/* the marked box with axes (for feedback) */
makeobj( wbox_obj );
  /* do the edges of the box */
  move( -1.0, -1.0, -1.0 );
  draw( -1.0,  1.0, -1.0 );
    passthrough( TOKTYP_ROT | TOKCOD_Y | TOKPT(0,0) );
  draw(  1.0,  1.0, -1.0 );
    passthrough( TOKTYP_ROT | TOKCOD_X | TOKPT(1,0) );
  draw(  1.0, -1.0, -1.0 );
    passthrough( TOKTYP_ROT | TOKCOD_Y | TOKPT(0,1) );
  draw( -1.0, -1.0, -1.0 );
    passthrough( TOKTYP_ROT | TOKCOD_X | TOKPT(0,0) );
  move( -1.0, -1.0,  1.0 );
  draw( -1.0,  1.0,  1.0 );
    passthrough( TOKTYP_ROT | TOKCOD_Y | TOKPT(1,0) );
  draw(  1.0,  1.0,  1.0 );
    passthrough( TOKTYP_ROT | TOKCOD_X | TOKPT(1,1) );
  draw(  1.0, -1.0,  1.0 );
    passthrough( TOKTYP_ROT | TOKCOD_Y | TOKPT(1,1) );
  draw( -1.0, -1.0,  1.0 );
    passthrough( TOKTYP_ROT | TOKCOD_X | TOKPT(0,1) );
  move( -1.0, -1.0, -1.0 );
  draw( -1.0, -1.0,  1.0 );
    passthrough( TOKTYP_ROT | TOKCOD_Z | TOKPT(0,0) );
  move( -1.0,  1.0, -1.0 );
  draw( -1.0,  1.0,  1.0 );
    passthrough( TOKTYP_ROT | TOKCOD_Z | TOKPT(0,1) );
  move(  1.0, -1.0, -1.0 );
  draw(  1.0, -1.0,  1.0 );
    passthrough( TOKTYP_ROT | TOKCOD_Z | TOKPT(1,0) );
  move(  1.0,  1.0, -1.0 );
  draw(  1.0,  1.0,  1.0 );
    passthrough( TOKTYP_ROT | TOKCOD_Z | TOKPT(1,1) );

#if 0
  /* do the axes */
  move( 0.0, 0.0, 0.0 );
  draw( 1.5, 0.0, 0.0 );
    passthrough( TOK_XAXIS );
  move( 0.0, 0.0, 0.0 );
  draw( 0.0, 1.5, 0.0 );
    passthrough( TOK_YAXIS );
  move( 0.0, 0.0, 0.0 );
  draw( 0.0, 0.0, 1.5 );
    passthrough( TOK_ZAXIS );
#endif
closeobj();
}



/*  ************************************
    *                                  *
    *        gen_wbox                  *
    *                                  *
    ************************************

    DESCRIPTION:

    ENTRY:
*/

void gen_wbox ( )
{
callobj( wbox_obj );
}




/*  ************************************
    *                                  *
    *      draw_rot                    *
    *                                  *
    ************************************

    DESCRIPTION:  Draw the rotation structure.

    ENTRY:

    GLOBALS:
      trysplat -- TRUE if we should check to draw the splat object
      trycell -- TRUE if we should check to draw the last picked cell
      lastpick -- token of last picked cell
*/

void draw_rot ( vptr )
VIEW *vptr;
{
use_view( vptr );
cm_color( seg1, s1_BACKG );
clear();

/* draw the splat */
#if DRAWSPLAT
  if ( trysplat  &&  getbutton(LEFTSHIFTKEY) )
    drawsplat();
#endif

/* draw the picked cell */
#if DRAWCELL
  if ( trycell  &&  getbutton(RIGHTSHIFTKEY) )
    drawcell( lastpick-1 );
#endif

draw_obj();
draw_box();
draw_xh( vptr );
}
