/*  FILENAME:  xgeom.c

    DESCRIPTION:  Transform a geom file.

    AUTHOR:  Kelvin Thompson, 4/25/89
*/

#include <stdio.h>
#include <math.h>

/* major types */
typedef double FLOAT;
typedef long INT;
typedef int BOOL;
typedef FLOAT XFORM[4][4];

/* programs in this file */
extern main();				/* entry point */
extern void read_xfile();		/* read the transform file */
extern struct SHELL *read_gfile();	/* read the geometry file */
extern struct SHELL *xform_gobj();	/* transform an object */
extern BOOL write_gfile();		/* write out geometry file */
extern void printuse();			/* print help message */
extern BOOL read3vals();		/* read three floats from a string */
extern void scale();			/* append scale op to matrix */
extern void move();			/* append translate op to matrix */
extern void rotate();			/* append rotate op to matrix */
extern XFORM *matxmul();		/* multiply matricies */

#ifndef NULL
#  define NULL 0
#endif

#ifndef TRUE
#  define TRUE 1
#  define FALSE 0
#endif

#define PI      3.1415926535897932384626433832795028841971693993751
#define DEGCOS(deg)  cos( (deg)*(PI/180.0) )
#define DEGSIN(deg)  sin( (deg)*(PI/180.0) )

typedef struct VEC3 { FLOAT x,y,z; } VEC3;

typedef struct POLYGON
  {
  int npoints;
  VEC3 **point;
  } POLYGON;

typedef struct SHELL
  {
  int npoints;
  int npolys;
  int nedges;
  VEC3 *point;
  POLYGON *poly;
  } SHELL;

typedef struct XHOLDER { XFORM x; } XHOLDER;

/* line buffer */
#define LINEBUFFSIZE 128
static char linebuff[LINEBUFFSIZE];

XHOLDER I =
  {{
  {1,0,0,0},
  {0,1,0,0},
  {0,0,1,0},
  {0,0,0,1}
  }};

/* macros to initialize and assign matricies */
#define INIT_XFORM(xf)       ( (*(XHOLDER *)(xf)) = I )
#define ASSIGN_XFORM(x1,x2)  ( (*(XHOLDER *)(x1)) = (*(XHOLDER *)(x2)) )

#define SKIP_TO_WHITE(cptr)                    \
  do {                                         \
     if ( *cptr == ' '  ||  *cptr == '\t'  ||  \
          *cptr == '\0'  ||  *cptr == '\n' )   \
       break;                                  \
     ++cptr;                                   \
     } while ( 1 )



/*  *******************************
    *                             *
    *        main                 *
    *                             *
    *******************************

    DESCRIPTION:

    ENTRY:
*/

main ( argc, argv )
int argc;
char **argv;
{
FILE *gfile;
XFORM xf;
SHELL *gobj;

/* check input line */
if ( argc < 3 )
  { printuse( stderr, "Too few parameters.\n" );  exit( 1 ); }
if ( argc > 3 )
  { printuse( stderr, "Too many parameters.\n" );  exit( 1 ); }

/* get transform from xform-file */
read_xfile( argv[1], xf );

/* open geom file */
if ( argv[2][0] == '-'  &&  argv[2][1] == '\0' )
  {
  /* use stdin */
  gfile = stdin;
  }
else
  {
  gfile = fopen( argv[2], "r" );
  if ( gfile == NULL )
    { printuse( stderr, "Can't open geom-file.\n" );  exit( 1 ); }
  }

/* read in shell object */
gobj = read_gfile( gfile );
if ( gobj == NULL )
  {
  fprintf( stderr, "Failed to read geom-file.\n" );
  exit( 1 );
  }

/* close geom-file */
if ( gfile != stdin )
  fclose( gfile );

/* apply the total transform */
xform_gobj( xf, gobj );

/* write transformed object to stdout */
if ( ! write_gfile(stdout,gobj) )
  {
  perror( "xgeom" );
  fprintf( stderr, "Failed to write geom-file.\n" );
  exit( 1 );
  }

exit( 0 );
}



/*  *******************************
    *                             *
    *       read_xfile            *
    *                             *
    *******************************

    DESCRIPTION:

    ENTRY:
*/

void read_xfile ( filename, xf )
char *filename;
XFORM *xf;
{
FILE *xfile;
char *retval;
FLOAT xval,yval,zval;
int linenum,rcount;
BOOL readok;

/* open xform file */
if ( filename[0] == '-'  &&  filename[1] == '\0' )  /* i.e. "-" */
  {
  /* use stdin */
  xfile = stdin;
  }
else
  {
  /* open the file */
  xfile = fopen( filename, "r" );
  if ( xfile == NULL )
    { printuse( stderr, "Can't open xform-file.\n" );  exit( 1 ); }
  }

/* init */
INIT_XFORM( xf );
linenum = 0;
readok = TRUE;

do
  {
  /* get next line */
  linenum++;
  retval = fgets( linebuff, LINEBUFFSIZE, xfile );
  if ( retval == NULL )
    break;

  switch ( linebuff[0] )
    {
    case 's':
    case 'S':
      if ( ! (readok = read3vals(&linebuff[1],&xval,&yval,&zval)) )
        break;
      scale( xf, xval, yval, zval );
      break;

    case 't':
    case 'T':
    case 'm':
    case 'M':
      if ( ! (readok = read3vals(&linebuff[1],&xval,&yval,&zval)) )
        break;
      move( xf, xval, yval, zval );
      break;

    case 'r':
    case 'R':
      if ( ! (readok = read3vals(&linebuff[1],&xval,&yval,&zval)) )
        break;
      rotate( xf, xval, yval, zval );
      break;
    }
  }  while ( retval != NULL  &&  readok );

if ( xfile != stdin )
  close( xfile );
}




/*  *******************************
    *                             *
    *      read_gfile             *
    *                             *
    *******************************

    DESCRIPTION:

    ENTRY:

    EXIT:
*/

struct SHELL *read_gfile ( fptr )
FILE *fptr;
{
register i,j;
register SHELL *obj;
VEC3 *pt;
POLYGON *poly;
int pointcount,polycount,vcount;
FLOAT ftemp,maxr;
int itemp;

/* get point and polygon counts */
fscanf( fptr, "%d %d %d", &pointcount, &polycount, &itemp );

/* allocate space */
obj = (SHELL *) malloc( sizeof(SHELL)
         + pointcount * sizeof(VEC3)
         + polycount * sizeof(POLYGON) );
if ( obj == NULL )
  {
  fprintf( stderr, "xgeom: Can't allocate space.\n" );
  exit( 1 );
  }

/* init object struct */
obj->npoints = pointcount;
obj->npolys = polycount;
obj->nedges = itemp;
obj->point = (VEC3 *) (& obj[1]);
obj->poly = (POLYGON *) (& obj->point[pointcount]);

/* read in points */
maxr = 0.0;
for (
    i = pointcount,   pt = obj->point;
    i;
    --i,              ++pt
    )
  {
  fscanf( fptr, "%lf %lf %lf", &pt->x, &pt->y, &pt->z );

  ftemp = fabs(pt->x);  if ( maxr < ftemp )  maxr = ftemp;
  ftemp = fabs(pt->y);  if ( maxr < ftemp )  maxr = ftemp;
  ftemp = fabs(pt->z);  if ( maxr < ftemp )  maxr = ftemp;
  }

/* read in polygons */
for (
    i = polycount,   poly = obj->poly;
    i;
    --i,              ++poly
    )
  {
  fscanf( fptr, "%d", &vcount );
  poly->npoints = vcount;

  poly->point = (VEC3 **) malloc( vcount * sizeof(VEC3 *) );
  if ( poly->point == NULL )
    {
    fprintf( stderr, "xgeom: Can't alloc tiny space.\n" );
    exit( 1 );
    }

  for ( j=0;  j<vcount;  ++j )
    {
    fscanf( fptr, "%d", &itemp );
    poly->point[j] = & obj->point[itemp-1];
    }
  }

return obj;
}



/*  *******************************
    *                             *
    *       xform_gobj            *
    *                             *
    *******************************

    DESCRIPTION:

    ENTRY:
*/

SHELL *xform_gobj( xf, gobj )
register XFORM *xf;
SHELL *gobj;
{
register i;
register VEC3 *pt;
VEC3 holdpt;

/* loop through points */
for (
    i = gobj->npoints,   pt = gobj->point;
    i;
    --i,                 ++pt
    )
  {
  holdpt.x = pt->x * (*xf)[0][0]
	   + pt->y * (*xf)[1][0]
	   + pt->z * (*xf)[2][0]
	   +         (*xf)[3][0];
  holdpt.y = pt->x * (*xf)[0][1]
	   + pt->y * (*xf)[1][1]
	   + pt->z * (*xf)[2][1]
	   +         (*xf)[3][1];
  holdpt.z = pt->x * (*xf)[0][2]
	   + pt->y * (*xf)[1][2]
	   + pt->z * (*xf)[2][2]
	   +         (*xf)[3][2];
  *pt = holdpt;
  }

return gobj;
}




/*  *******************************
    *                             *
    *    write_gfile              *
    *                             *
    *******************************

    DESCRIPTION:

    ENTRY:
*/

BOOL write_gfile ( fptr, gobj )
FILE *fptr;
register SHELL *gobj;
{
register i,j;
VEC3 *pt;
POLYGON *poly;
int vcount;
int retval;

/* write point and polygon counts */
retval = fprintf( fptr, "%d %d %d\n",
  gobj->npoints, gobj->npolys, gobj->nedges );
if ( retval == EOF ) return FALSE;

/* write out points */
for (
    i = gobj->npoints,  pt = gobj->point;
    i;
    --i,                ++pt
    )
  {
  retval = fprintf( fptr, "%22.17lg %22.17lg %22.17lg\n",
    pt->x, pt->y, pt->z );
  if ( retval == EOF ) return FALSE;
  }

/* write out polygons */
for (
    i = gobj->npolys,  poly = gobj->poly;
    i;
    --i,               ++poly
    )
  {
  vcount = poly->npoints;
  fprintf( fptr, "%d", vcount );

  for ( j=0;  j<vcount;  ++j )
    {
    retval = fprintf( fptr, " %d", (poly->point[j] - gobj->point) + 1 );
    if ( retval == EOF ) return FALSE;
    }

  putc( '\n', fptr );
  }

return TRUE;
}



/*  *******************************
    *                             *
    *       printuse              *
    *                             *
    *******************************

    DESCRIPTION:  Print usage information.

    ENTRY:
      stream -- output stream
      message -- string containing message
*/

void printuse ( stream, message )
FILE *stream;
char *message;
{
if ( message != NULL  &&  message[0] != '\0' )
  fputs( message, stream );

fprintf( stream, "Usage: xgeom xform-file geom-file\n" );
}



/*  *******************************
    *                             *
    *      read3vals              *
    *                             *
    *******************************

    DESCRIPTION:

    ENTRY:
*/

BOOL read3vals ( cptr, xval, yval, zval )
register char *cptr;
FLOAT *xval,*yval,*zval;
{
int rcount;

/* move to parameters */
SKIP_TO_WHITE( cptr );
if ( *cptr == '\0'  ||  *cptr == '\n' )
  return FALSE;

/* read parameters */
rcount = sscanf( cptr, "%lf %lf %lf", xval, yval, zval );
if ( rcount != 3 )
  return FALSE;

return TRUE;
}



/*  *******************************
    *                             *
    *         scale               *
    *                             *
    *******************************

    DESCRIPTION:

    ENTRY:
*/

void scale ( xf, xval, yval, zval )
XFORM *xf;
FLOAT xval,yval,zval;
{
XFORM scale_xf;

INIT_XFORM( scale_xf );

scale_xf[0][0] = xval;
scale_xf[1][1] = yval;
scale_xf[2][2] = zval;

matxmul( xf, xf, scale_xf );
}



/*  *******************************
    *                             *
    *          move               *
    *                             *
    *******************************

    DESCRIPTION:

    ENTRY:
*/

void move ( xf, xval, yval, zval )
XFORM *xf;
FLOAT xval,yval,zval;
{
XFORM move_xf;

INIT_XFORM( move_xf );

move_xf[3][0] = xval;
move_xf[3][1] = yval;
move_xf[3][2] = zval;

matxmul( xf, xf, move_xf );
}




/*  *******************************
    *                             *
    *         rotate              *
    *                             *
    *******************************

    DESCRIPTION:

    ENTRY:
*/

void rotate ( xf, xval, yval, zval )
XFORM *xf;
FLOAT xval,yval,zval;
{
register FLOAT cosval,sinval;
XFORM rot_xf;

if ( xval != 0.0 )
  {
  cosval = DEGCOS( xval );
  sinval = DEGSIN( xval );

  INIT_XFORM( rot_xf );
  rot_xf[1][1] =  cosval;   rot_xf[1][2] =  sinval;
  rot_xf[2][1] = -sinval;   rot_xf[2][2] =  cosval;
  matxmul( xf, xf, rot_xf );
  }

if ( yval != 0.0 )
  {
  cosval = DEGCOS( yval );
  sinval = DEGSIN( yval );

  INIT_XFORM( rot_xf );
  rot_xf[0][0] =  cosval;   rot_xf[0][2] = -sinval;
  rot_xf[2][0] =  sinval;   rot_xf[2][2] =  cosval;
  matxmul( xf, xf, rot_xf );
  }

if ( zval != 0.0 )
  {
  cosval = DEGCOS( zval );
  sinval = DEGSIN( zval );

  INIT_XFORM( rot_xf );
  rot_xf[0][0] =  cosval;   rot_xf[0][1] =  sinval;
  rot_xf[1][0] = -sinval;   rot_xf[1][1] =  cosval;
  matxmul( xf, xf, rot_xf );
  }
}




/*  ************************************
    *                                  *
    *           matxmul                *
    *                                  *
    ************************************

    DESCRIPTION:  Multiply two 4x4 matricies.  Assume the last
      column of matrix 'b' is {0,0,0,1}.

    ENTRY:

    EXIT:
*/

XFORM *matxmul ( result, a, b )
register XFORM *result,*a,*b;
{
XFORM tempx;

tempx[0][0] = (*a)[0][0] * (*b)[0][0]
	    + (*a)[0][1] * (*b)[1][0]
	    + (*a)[0][2] * (*b)[2][0]
	    + (*a)[0][3] * (*b)[3][0];

tempx[0][1] = (*a)[0][0] * (*b)[0][1]
	    + (*a)[0][1] * (*b)[1][1]
	    + (*a)[0][2] * (*b)[2][1]
	    + (*a)[0][3] * (*b)[3][1];

tempx[0][2] = (*a)[0][0] * (*b)[0][2]
	    + (*a)[0][1] * (*b)[1][2]
	    + (*a)[0][2] * (*b)[2][2]
	    + (*a)[0][3] * (*b)[3][2];

tempx[0][3] = (*a)[0][0] * (*b)[0][3]
	    + (*a)[0][1] * (*b)[1][3]
	    + (*a)[0][2] * (*b)[2][3]
	    + (*a)[0][3] * (*b)[3][3];

tempx[1][0] = (*a)[1][0] * (*b)[0][0]
	    + (*a)[1][1] * (*b)[1][0]
	    + (*a)[1][2] * (*b)[2][0]
	    + (*a)[1][3] * (*b)[3][0];

tempx[1][1] = (*a)[1][0] * (*b)[0][1]
	    + (*a)[1][1] * (*b)[1][1]
	    + (*a)[1][2] * (*b)[2][1]
	    + (*a)[1][3] * (*b)[3][1];

tempx[1][2] = (*a)[1][0] * (*b)[0][2]
	    + (*a)[1][1] * (*b)[1][2]
	    + (*a)[1][2] * (*b)[2][2]
	    + (*a)[1][3] * (*b)[3][2];

tempx[1][3] = (*a)[1][0] * (*b)[0][3]
	    + (*a)[1][1] * (*b)[1][3]
	    + (*a)[1][2] * (*b)[2][3]
	    + (*a)[1][3] * (*b)[3][3];

tempx[2][0] = (*a)[2][0] * (*b)[0][0]
	    + (*a)[2][1] * (*b)[1][0]
	    + (*a)[2][2] * (*b)[2][0]
	    + (*a)[2][3] * (*b)[3][0];

tempx[2][1] = (*a)[2][0] * (*b)[0][1]
	    + (*a)[2][1] * (*b)[1][1]
	    + (*a)[2][2] * (*b)[2][1]
	    + (*a)[2][3] * (*b)[3][1];

tempx[2][2] = (*a)[2][0] * (*b)[0][2]
	    + (*a)[2][1] * (*b)[1][2]
	    + (*a)[2][2] * (*b)[2][2]
	    + (*a)[2][3] * (*b)[3][2];

tempx[2][3] = (*a)[2][0] * (*b)[0][3]
	    + (*a)[2][1] * (*b)[1][3]
	    + (*a)[2][2] * (*b)[2][3]
	    + (*a)[2][3] * (*b)[3][3];

tempx[3][0] = (*a)[3][0] * (*b)[0][0]
	    + (*a)[3][1] * (*b)[1][0]
	    + (*a)[3][2] * (*b)[2][0]
	    + (*a)[3][3] * (*b)[3][0];

tempx[3][1] = (*a)[3][0] * (*b)[0][1]
	    + (*a)[3][1] * (*b)[1][1]
	    + (*a)[3][2] * (*b)[2][1]
	    + (*a)[3][3] * (*b)[3][1];

tempx[3][2] = (*a)[3][0] * (*b)[0][2]
	    + (*a)[3][1] * (*b)[1][2]
	    + (*a)[3][2] * (*b)[2][2]
	    + (*a)[3][3] * (*b)[3][2];

tempx[3][3] = (*a)[3][0] * (*b)[0][3]
	    + (*a)[3][1] * (*b)[1][3]
	    + (*a)[3][2] * (*b)[2][3]
	    + (*a)[3][3] * (*b)[3][3];

ASSIGN_XFORM( result, tempx );
return result;
}
